package com.sun.media.sound;

import javax.sound.sampled.AudioInputStream;
import java.io.InputStream;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineUnavailableException;
import java.io.IOException;
import java.util.Arrays;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.SourceDataLine;

public final class SoftMixingSourceDataLine extends SoftMixingDataLine implements SourceDataLine
{
    private boolean open;
    private AudioFormat format;
    private int framesize;
    private int bufferSize;
    private float[] readbuffer;
    private boolean active;
    private byte[] cycling_buffer;
    private int cycling_read_pos;
    private int cycling_write_pos;
    private int cycling_avail;
    private long cycling_framepos;
    private AudioFloatInputStream afis;
    private boolean _active;
    private AudioFormat outputformat;
    private int out_nrofchannels;
    private int in_nrofchannels;
    private float _rightgain;
    private float _leftgain;
    private float _eff1gain;
    private float _eff2gain;
    
    SoftMixingSourceDataLine(SoftMixingMixer softMixingMixer, DataLine.Info info) {
        super(softMixingMixer, info);
        this.open = false;
        this.format = new AudioFormat(44100.0f, 16, 2, true, false);
        this.bufferSize = -1;
        this.active = false;
        this.cycling_read_pos = 0;
        this.cycling_write_pos = 0;
        this.cycling_avail = 0;
        this.cycling_framepos = 0L;
        this._active = false;
    }
    
    @Override
    public int write(byte[] array, int index, int n) {
        if (!this.isOpen()) {
            return 0;
        }
        if (n % this.framesize != 0) {
            throw new IllegalArgumentException("Number of bytes does not represent an integral number of sample frames.");
        }
        if (index < 0) {
            throw new ArrayIndexOutOfBoundsException(index);
        }
        if ((long)index + (long)n > (long)array.length) {
            throw new ArrayIndexOutOfBoundsException(array.length);
        }
        byte[] cycling_buffer = this.cycling_buffer;
        int length = this.cycling_buffer.length;
        int i = 0;
        while (i != n) {
            int cycling_avail;
            synchronized (this.cycling_buffer) {
                int cycling_write_pos = this.cycling_write_pos;
                cycling_avail = this.cycling_avail;
                while (i != n && cycling_avail != length) {
                    cycling_buffer[cycling_write_pos++] = array[index++];
                    ++i;
                    ++cycling_avail;
                    if (cycling_write_pos != length) {
                        continue;
                    }
                    cycling_write_pos = 0;
                }
                this.cycling_avail = cycling_avail;
                this.cycling_write_pos = cycling_write_pos;
                if (i == n) {
                    return i;
                }
            }
            if (cycling_avail == length) {
                try {
                    Thread.sleep(1L);
                }
                catch (InterruptedException ex) {
                    return i;
                }
                if (!this.isRunning()) {
                    return i;
                }
                continue;
            }
            else {
                continue;
            }
        }
        return i;
    }
    
    @Override
    protected void processControlLogic() {
        this._active = this.active;
        this._rightgain = this.rightgain;
        this._leftgain = this.leftgain;
        this._eff1gain = this.eff1gain;
        this._eff2gain = this.eff2gain;
    }
    
    @Override
    protected void processAudioLogic(SoftAudioBuffer[] array) {
        if (this._active) {
            float[] array2 = array[0].array();
            float[] array3 = array[1].array();
            int size = array[0].getSize();
            int toIndex = size * this.in_nrofchannels;
            if (this.readbuffer == null || this.readbuffer.length < toIndex) {
                this.readbuffer = new float[toIndex];
            }
            try {
                int read = this.afis.read(this.readbuffer);
                if (read != this.in_nrofchannels) {
                    Arrays.fill(this.readbuffer, read, toIndex, 0.0f);
                }
            }
            catch (IOException ex) {}
            int in_nrofchannels = this.in_nrofchannels;
            for (int i = 0, n = 0; i < size; ++i, n += in_nrofchannels) {
                int n2 = i;
                array2[n2] += this.readbuffer[n] * this._leftgain;
            }
            if (this.out_nrofchannels != 1) {
                if (this.in_nrofchannels == 1) {
                    for (int k = 0, n5 = 0; k < size; ++k, n5 += in_nrofchannels) {
                        int n6 = k;
                        array3[n6] += this.readbuffer[n5] * this._rightgain;
                    }
                }
                else {
                    for (int j = 0, n3 = 1; j < size; ++j, n3 += in_nrofchannels) {
                        int n4 = j;
                        array3[n4] += this.readbuffer[n3] * this._rightgain;
                    }
                }
            }
            if ((double)this._eff1gain > 1.0E-4) {
                float[] array4 = array[2].array();
                for (int l = 0, n7 = 0; l < size; ++l, n7 += in_nrofchannels) {
                    int n8 = l;
                    array4[n8] += this.readbuffer[n7] * this._eff1gain;
                }
                if (this.in_nrofchannels == 2) {
                    for (int n9 = 0, n10 = 1; n9 < size; ++n9, n10 += in_nrofchannels) {
                        int n11 = n9;
                        array4[n11] += this.readbuffer[n10] * this._eff1gain;
                    }
                }
            }
            if ((double)this._eff2gain > 1.0E-4) {
                float[] array5 = array[3].array();
                for (int n12 = 0, n13 = 0; n12 < size; ++n12, n13 += in_nrofchannels) {
                    int n14 = n12;
                    array5[n14] += this.readbuffer[n13] * this._eff2gain;
                }
                if (this.in_nrofchannels == 2) {
                    for (int n15 = 0, n16 = 1; n15 < size; ++n15, n16 += in_nrofchannels) {
                        int n17 = n15;
                        array5[n17] += this.readbuffer[n16] * this._eff2gain;
                    }
                }
            }
        }
    }
    
    @Override
    public void open() throws LineUnavailableException {
        this.open(this.format);
    }
    
    @Override
    public void open(AudioFormat audioFormat) throws LineUnavailableException {
        if (this.bufferSize == -1) {
            this.bufferSize = (int)(audioFormat.getFrameRate() / 2.0f) * audioFormat.getFrameSize();
        }
        this.open(audioFormat, this.bufferSize);
    }
    
    @Override
    public void open(AudioFormat audioFormat, int n) throws LineUnavailableException {
        Object o = null;
        if (n < audioFormat.getFrameSize() * 32) {
            n = audioFormat.getFrameSize() * 32;
        }
        synchronized (this.control_mutex) {
            if (!this.isOpen()) {
                if (!this.mixer.isOpen()) {
                    this.mixer.open();
                    this.mixer.implicitOpen = true;
                }
                o = new LineEvent(this, LineEvent.Type.OPEN, 0L);
                this.bufferSize = n - n % audioFormat.getFrameSize();
                this.format = audioFormat;
                this.framesize = audioFormat.getFrameSize();
                this.outputformat = this.mixer.getFormat();
                this.out_nrofchannels = this.outputformat.getChannels();
                this.in_nrofchannels = audioFormat.getChannels();
                this.open = true;
                this.mixer.getMainMixer().openLine(this);
                this.cycling_buffer = new byte[this.framesize * n];
                this.cycling_read_pos = 0;
                this.cycling_write_pos = 0;
                this.cycling_avail = 0;
                this.cycling_framepos = 0L;
                this.afis = AudioFloatInputStream.getInputStream(new AudioInputStream(new InputStream() {
                    @Override
                    public int read() throws IOException {
                        byte[] b = { 0 };
                        int read = super.read(b);
                        if (read < 0) {
                            return read;
                        }
                        return b[0] & 0xFF;
                    }
                    
                    @Override
                    public int available() throws IOException {
                        synchronized (SoftMixingSourceDataLine.this.cycling_buffer) {
                            return SoftMixingSourceDataLine.this.cycling_avail;
                        }
                    }
                    
                    @Override
                    public int read(byte[] array, int n, int access$100) throws IOException {
                        synchronized (SoftMixingSourceDataLine.this.cycling_buffer) {
                            if (access$100 > SoftMixingSourceDataLine.this.cycling_avail) {
                                access$100 = SoftMixingSourceDataLine.this.cycling_avail;
                            }
                            int access$101 = SoftMixingSourceDataLine.this.cycling_read_pos;
                            byte[] access$102 = SoftMixingSourceDataLine.this.cycling_buffer;
                            int length = access$102.length;
                            for (int i = 0; i < access$100; ++i) {
                                array[n++] = access$102[access$101];
                                if (++access$101 == length) {
                                    access$101 = 0;
                                }
                            }
                            SoftMixingSourceDataLine.this.cycling_read_pos = access$101;
                            SoftMixingSourceDataLine.this.cycling_avail -= access$100;
                            SoftMixingSourceDataLine.this.cycling_framepos += (long)(access$100 / SoftMixingSourceDataLine.this.framesize);
                        }
                        return access$100;
                    }
                }, audioFormat, -1L));
                this.afis = new SoftMixingSourceDataLine.NonBlockingFloatInputStream(this.afis);
                if ((double)Math.abs(audioFormat.getSampleRate() - this.outputformat.getSampleRate()) > 1.0E-6) {
                    this.afis = new SoftMixingDataLine.AudioFloatInputStreamResampler(this.afis, this.outputformat);
                }
            }
            else if (!audioFormat.matches(this.getFormat())) {
                throw new IllegalStateException("Line is already open with format " + this.getFormat() + " and bufferSize " + this.getBufferSize());
            }
        }
        if (o != null) {
            this.sendEvent(o);
        }
    }
    
    @Override
    public int available() {
        synchronized (this.cycling_buffer) {
            return this.cycling_buffer.length - this.cycling_avail;
        }
    }
    
    @Override
    public void drain() {
        while (true) {
            int cycling_avail;
            synchronized (this.cycling_buffer) {
                cycling_avail = this.cycling_avail;
            }
            if (cycling_avail != 0) {
                break;
            }
            try {
                Thread.sleep(1L);
            }
            catch (InterruptedException ex) {
                return;
            }
        }
    }
    
    @Override
    public void flush() {
        synchronized (this.cycling_buffer) {
            this.cycling_read_pos = 0;
            this.cycling_write_pos = 0;
            this.cycling_avail = 0;
        }
    }
    
    @Override
    public int getBufferSize() {
        synchronized (this.control_mutex) {
            return this.bufferSize;
        }
    }
    
    @Override
    public AudioFormat getFormat() {
        synchronized (this.control_mutex) {
            return this.format;
        }
    }
    
    @Override
    public int getFramePosition() {
        return (int)this.getLongFramePosition();
    }
    
    @Override
    public float getLevel() {
        return -1.0f;
    }
    
    @Override
    public long getLongFramePosition() {
        synchronized (this.cycling_buffer) {
            return this.cycling_framepos;
        }
    }
    
    @Override
    public long getMicrosecondPosition() {
        return (long)((double)this.getLongFramePosition() * (1000000.0 / (double)this.getFormat().getSampleRate()));
    }
    
    @Override
    public boolean isActive() {
        synchronized (this.control_mutex) {
            return this.active;
        }
    }
    
    @Override
    public boolean isRunning() {
        synchronized (this.control_mutex) {
            return this.active;
        }
    }
    
    @Override
    public void start() {
        LineEvent lineEvent = null;
        synchronized (this.control_mutex) {
            if (this.isOpen()) {
                if (this.active) {
                    return;
                }
                this.active = true;
                lineEvent = new LineEvent(this, LineEvent.Type.START, this.getLongFramePosition());
            }
        }
        if (lineEvent != null) {
            this.sendEvent(lineEvent);
        }
    }
    
    @Override
    public void stop() {
        LineEvent lineEvent = null;
        synchronized (this.control_mutex) {
            if (this.isOpen()) {
                if (!this.active) {
                    return;
                }
                this.active = false;
                lineEvent = new LineEvent(this, LineEvent.Type.STOP, this.getLongFramePosition());
            }
        }
        if (lineEvent != null) {
            this.sendEvent(lineEvent);
        }
    }
    
    @Override
    public void close() {
        LineEvent lineEvent = null;
        synchronized (this.control_mutex) {
            if (!this.isOpen()) {
                return;
            }
            this.stop();
            lineEvent = new LineEvent(this, LineEvent.Type.CLOSE, this.getLongFramePosition());
            this.open = false;
            this.mixer.getMainMixer().closeLine(this);
        }
        if (lineEvent != null) {
            this.sendEvent(lineEvent);
        }
    }
    
    @Override
    public boolean isOpen() {
        synchronized (this.control_mutex) {
            return this.open;
        }
    }
    
    private static class NonBlockingFloatInputStream extends AudioFloatInputStream
    {
        AudioFloatInputStream ais;
        
        NonBlockingFloatInputStream(AudioFloatInputStream ais) {
            this.ais = ais;
        }
        
        @Override
        public int available() throws IOException {
            return this.ais.available();
        }
        
        @Override
        public void close() throws IOException {
            this.ais.close();
        }
        
        @Override
        public AudioFormat getFormat() {
            return this.ais.getFormat();
        }
        
        @Override
        public long getFrameLength() {
            return this.ais.getFrameLength();
        }
        
        @Override
        public void mark(int n) {
            this.ais.mark(n);
        }
        
        @Override
        public boolean markSupported() {
            return this.ais.markSupported();
        }
        
        @Override
        public int read(float[] a, int n, int n2) throws IOException {
            int available = this.available();
            if (n2 > available) {
                Arrays.fill(a, n + this.ais.read(a, n, available), n + n2, 0.0f);
                return n2;
            }
            return this.ais.read(a, n, n2);
        }
        
        @Override
        public void reset() throws IOException {
            this.ais.reset();
        }
        
        @Override
        public long skip(long n) throws IOException {
            return this.ais.skip(n);
        }
    }
}
